/*
 * Local Matters, Inc. Confidential
 * OCO Source Materials
 * (C) Local Matters, Inc. 2007. All Rights Reserved.
 */
package com.localmatters.flexiq.collector.browser;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.annotation.Resource;
import javax.ejb.Stateful;
import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.Message;
import javax.jms.Queue;
import javax.jms.QueueBrowser;
import javax.jms.Session;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * The collector browser service implementation.
 *
 * @author Daniel Cavalcanti
 */
@Stateful()
public class BrowserBean
    implements BrowserService {
    
    // **********************************************************************
    // Instance variables
    // **********************************************************************
    
    // Logging facility *****************************************************
    
    /**
     * The class logger.
     */
    private Log log = LogFactory.getLog(getClass());
    
    // JMS objects **********************************************************
    
    @Resource(mappedName="jms/CollectorConnectionFactory")
    private ConnectionFactory connectionFactory;
    
    private Connection connection;
    private Session session;
    
    // Injected resources ***************************************************
    
    @Resource(name="username")
    private String username;
    
    @Resource(name="password")
    private String password;
    
    // Helper objects *******************************************************
    
    private InitialContext ic;
    
    // **********************************************************************
    // Constructors
    // **********************************************************************
    
    /**
     * Creates a new instance of BrowserBean.
     */
    public BrowserBean() {
        if (log.isTraceEnabled()) log.trace("Creating collector browser bean.");
    }
    
    // **********************************************************************
    // Life-cycle methods
    // **********************************************************************
    
    @PostConstruct()
    private void init() {
        
        if (log.isTraceEnabled()) log.trace("Initializing collector browser bean.");
        
        // ******************************************************************
        // Establish JMS connectivity
        // ******************************************************************
        
        try {
            
            if (log.isTraceEnabled()) log.trace("Getting server initial context.");
            ic = new InitialContext();
            
            if (connectionFactory == null) {
                if (log.isTraceEnabled()) log.trace("Connection factory not set. Retrieving from server context.");
                connectionFactory = (ConnectionFactory) ic.lookup("jms/CollectorConnectionFactory");
            }
            
            if (log.isTraceEnabled()) log.trace("Establishing connection to JMS broker.");
            connection = connectionFactory.createConnection(username, password);
            
            if (log.isTraceEnabled()) log.trace("Creating transacted session with JMS broker.");
            session = connection.createSession(true, Session.SESSION_TRANSACTED);
            
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
        }
        
        if (log.isTraceEnabled()) log.trace("Collector browser bean started.");
        
    }
    
    @PreDestroy()
    private void release() {
        
        try {
            if (connection != null) {
                if (log.isTraceEnabled()) log.trace("Closing JMS connection.");
                connection.close();
            }
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
        }
        
        if (log.isTraceEnabled()) log.trace("Collector browser stopped.");
        
    }
    
    // **********************************************************************
    // BrowserService methods
    // **********************************************************************
    
    public List<Message> getMessages(String queueJndiName, String username, String password)
    throws BrowserException {
        
        validateCredentials(username, password);
        
        try {
            
            QueueBrowser browser = session.createBrowser(getQueue(queueJndiName));
            List<Message> messages = new ArrayList<Message>();
            
            for (Enumeration enumeration = browser.getEnumeration() ; enumeration.hasMoreElements() ; )
                messages.add((Message) enumeration.nextElement());
            
            browser.close();
            
            return messages;
            
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
            throw new BrowserException("Unexpected error.", ex);
        }
        
    }
    
    public List<Message> getMessages(String queueJndiName, String username, String password, int maxMessages)
    throws BrowserException {
        
        validateCredentials(username, password);
        
        try {
            
            QueueBrowser browser = session.createBrowser(getQueue(queueJndiName));
            List<Message> messages = new ArrayList<Message>();
            
            int i = 0;
            for (Enumeration enumeration = browser.getEnumeration() ; enumeration.hasMoreElements() && i < maxMessages ; i++)
                messages.add((Message) enumeration.nextElement());
            
            browser.close();
            
            return messages;
            
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
            throw new BrowserException("Unexpected error.", ex);
        }
        
    }
    
    // **********************************************************************
    // Auxiliary methods
    // **********************************************************************
    
    private Queue getQueue(String queueJndiName)
    throws NamingException {
        return (Queue) ic.lookup(queueJndiName);
    }
    
    private void validateCredentials(String username, String password)
    throws BrowserException {
        if (!username.equals(this.username))
            throw new BrowserException("Invalid username.");
        if (!password.equals(this.password))
            throw new BrowserException("Invalid password.");
    }
    
}
