/*
 * Local Matters, Inc. Confidential
 * OCO Source Materials
 * (C) Local Matters, Inc. 2007. All Rights Reserved.
 */
package com.localmatters.flexiq.collector.consumer;

import static com.localmatters.flexiq.collector.CollectorConstants.*;
import com.localmatters.flexiq.security.CipherAlgorithmEnum;
import com.localmatters.flexiq.security.CipherService;
import com.localmatters.flexiq.security.CipherServiceBase;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.util.ArrayList;
import java.util.List;
import java.util.zip.DataFormatException;
import java.util.zip.Inflater;
import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.annotation.Resource;
import javax.ejb.Stateless;
import javax.ejb.TransactionManagement;
import javax.ejb.TransactionManagementType;
import javax.jms.BytesMessage;
import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.MessageConsumer;
import javax.jms.Session;
import javax.jws.WebMethod;
import javax.jws.WebParam;
import javax.jws.WebService;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.transaction.UserTransaction;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * The collector message consumer implementation.
 *
 * @author Daniel Cavalcanti
 */
@Stateless()
@WebService()
//@TransactionManagement(TransactionManagementType.BEAN)
public class ConsumerWS
    implements ConsumerService {
    
    // **********************************************************************
    // Instance variables
    // **********************************************************************
    
    // Logging facility *****************************************************
    
    /**
     * The class logger.
     */
    private Log log = LogFactory.getLog(getClass());
    
    // JMS objects **********************************************************
    
    @Resource(mappedName="jms/CollectorConnectionFactory")
    private ConnectionFactory connectionFactory;
    
    private Connection connection;
    private Session session;
    
    @Resource()
    private UserTransaction utx;
    
    // Injected resources ***************************************************
    
    @Resource(name="username")
    private String username;
    
    @Resource(name="password")
    private String password;
    
    // Helper objects *******************************************************
    
    private InitialContext ic;
    private Hex hexCodec = new Hex();
    private Inflater decompressor;
    
    // **********************************************************************
    // Constructors
    // **********************************************************************
    
    /**
     * Creates a new instance of ConsumerWS.
     */
    public ConsumerWS() {
        if (log.isTraceEnabled()) log.trace("Creating collector consumer bean.");
    }
    
    // **********************************************************************
    // Life-cycle methods
    // **********************************************************************
    
    /**
     * Initialize resources.
     */
    @PostConstruct()
    private void init() {
        
        if (log.isTraceEnabled()) log.trace("Initializing collector consumer bean.");
        
        // ******************************************************************
        // Establish JMS connectivity
        // ******************************************************************
        
        try {
            
            if (log.isTraceEnabled()) log.trace("Getting server initial context.");
            ic = new InitialContext();
            
            if (connectionFactory == null) {
                if (log.isTraceEnabled()) log.trace("Connection factory not set. Retrieving from server context.");
                connectionFactory = (ConnectionFactory) ic.lookup("jms/CollectorConnectionFactory");
            }
            
            if (log.isTraceEnabled()) log.trace("Establishing connection to JMS broker.");
            connection = connectionFactory.createConnection(username, password);
            
            if (log.isTraceEnabled()) log.trace("Creating transacted session with JMS broker.");
            session = connection.createSession(true, Session.SESSION_TRANSACTED);
//            session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
            
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
        }
        
        // ******************************************************************
        // Message processors
        // ******************************************************************
        
        try {
            
            if (log.isTraceEnabled()) log.trace("Initializing message decompressor.");
            decompressor = new Inflater();
            
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
        }
        
        if (log.isTraceEnabled()) log.trace("Collector consumer bean started.");
        
    }
    
    /**
     * Releases resources.
     */
    @PreDestroy()
    private void release() {
        
        try {
            if (connection != null) {
                if (log.isTraceEnabled()) log.trace("Closing JMS connection.");
                connection.close();
            }
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
        }
        
        if (log.isTraceEnabled()) log.trace("Collector consumer stopped.");
        
    }
    
    // **********************************************************************
    // ProducerService methods
    // **********************************************************************
    
    @WebMethod()
    public ResultMessage[] retrieve(@WebParam(name="username") String username, @WebParam(name="password") String password, @WebParam(name="jndiDestination") String jndiDestination, @WebParam(name="maxMessages") int maxMessages, @WebParam(name="compressed") boolean compressed)
    throws ConsumerException {
        
        // ******************************************************************
        // Validate parameters
        // ******************************************************************
        
        if (username == null)
            throw new ConsumerException(new NullPointerException("JMS connection username cannot be null."));
        if (password == null)
            throw new ConsumerException(new NullPointerException("JMS connection password cannot be null."));
        if (!this.username.equals(username))
            throw new ConsumerException(new IllegalArgumentException("Invalid JMS connection username."));
        if (!this.password.equals(password))
            throw new ConsumerException(new IllegalArgumentException("Invalid JMS connection password."));
        
        try {
            
//            utx.begin();
            
            // **************************************************************
            // Locate bound destination and create consumer
            // **************************************************************
            
            if (log.isTraceEnabled()) log.trace("Searching for named destination: " + jndiDestination);
            Destination destination = (Destination) ic.lookup(jndiDestination);
            
            if (log.isTraceEnabled()) log.trace("Creating consumer for named destination " + jndiDestination);
            MessageConsumer consumer = session.createConsumer(destination);
            
            if (log.isTraceEnabled()) log.trace("Starting JMS connection.");
            connection.start();
            
            // **************************************************************
            // Consume messages
            // **************************************************************
            
            if (log.isDebugEnabled()) log.trace("Creating retrieval containers.");
            List<ResultMessage> processedMessages = new ArrayList<ResultMessage>(maxMessages);
            BytesMessage jmsMessage = null;
            
            for (int i = 0 ; i < maxMessages ; i++) {
                
                // **********************************************************
                // Attempt message retrieve
                // **********************************************************
                
                if (log.isTraceEnabled()) log.trace("Attempting retrieval: " + i);
                jmsMessage = (BytesMessage) consumer.receive();
                
                // **********************************************************
                // Process retrieved message
                // **********************************************************
                
                if (jmsMessage != null) {
                    
                    if (log.isTraceEnabled()) log.trace("Message retrieved\n" + jmsMessage);
                    
                    // ******************************************************
                    // Extract message
                    // ******************************************************
                    
                    if (log.isTraceEnabled()) log.trace("Extracting result message container from JMS message.");
                    byte[] extracted = new byte[(int) jmsMessage.getBodyLength()];
                    jmsMessage.readBytes(extracted);
                    
                    // ******************************************************
                    // Decrypt message
                    // ******************************************************
                    
                    if (jmsMessage.propertyExists(ENCRYPT_HEADER) && jmsMessage.getBooleanProperty(ENCRYPTED_HEADER)) {
                        
                        if (log.isTraceEnabled()) log.trace("Loading cipher key from message.");
                        byte[] decodedEncryptionKey = hexCodec.decodeHex(jmsMessage.getStringProperty(ENCRYPTION_KEY).toCharArray());
                        
                        CipherService cipherService = null;
                        if (ENCRYPTION_ALGORITHM_AES.equals(jmsMessage.getStringProperty(ENCRYPTION_ALGORITHM)))
                            cipherService = CipherServiceBase.getInstance(CipherAlgorithmEnum.AES, decodedEncryptionKey);
                        else if (ENCRYPTION_ALGORITHM_BLOWFISH.equals(jmsMessage.getStringProperty(ENCRYPTION_ALGORITHM)))
                            cipherService = CipherServiceBase.getInstance(CipherAlgorithmEnum.Blowfish, decodedEncryptionKey);
                        else if (ENCRYPTION_ALGORITHM_DES.equals(jmsMessage.getStringProperty(ENCRYPTION_ALGORITHM)))
                            cipherService = CipherServiceBase.getInstance(CipherAlgorithmEnum.DES, decodedEncryptionKey);
                        else if (ENCRYPTION_ALGORITHM_TRIPLE_DES.equals(jmsMessage.getStringProperty(ENCRYPTION_ALGORITHM)))
                            cipherService = CipherServiceBase.getInstance(CipherAlgorithmEnum.TRIPLE_DES, decodedEncryptionKey);
                        else
                            throw new IllegalArgumentException("Invalid encryption algorithm: " + jmsMessage.getStringProperty(ENCRYPTION_ALGORITHM));
                        
                        if (log.isTraceEnabled()) log.trace("Decrypting message.");
                        extracted = decrypt(extracted, cipherService);
                        
                    }
                    
                    // ******************************************************
                    // Decompress message
                    // ******************************************************
                    
                    if (jmsMessage.getBooleanProperty(COMPRESSED_HEADER) && !compressed) {
                        if (log.isTraceEnabled()) log.trace("Decompressing message.");
                        extracted = decompress(extracted);
                    }
                    
                    // ******************************************************
                    // Done processing message
                    // ******************************************************
                    
                    if (log.isTraceEnabled()) log.trace("Message added to retrieval container.");
                    
                    String signature = jmsMessage.getStringProperty(DIGITAL_SIGNATURE);
                    processedMessages.add(new ResultMessage(extracted, signature));
                    
                } else
                    if (log.isTraceEnabled()) log.trace("No message was available.");
                
            }
            
//            if (log.isTraceEnabled()) log.trace("Closing consumer for named destination: " + jndiDestination);
//            consumer.close();
//
//            utx.commit();
            
            // **************************************************************
            // Package return container
            // **************************************************************
            
            if (log.isTraceEnabled()) log.trace("Packing retrieved messages to return.");
            ResultMessage[] collectorMessages = new ResultMessage[processedMessages.size()];
            for (int i = 0 ; i < collectorMessages.length ; i++)
                collectorMessages[i] = processedMessages.get(i);
            
            if (log.isTraceEnabled()) log.trace("Returning " + collectorMessages.length + " messages.");
            return collectorMessages;
            
        } catch (NamingException ex) {
            log.error("Unable to locate named queue: " + jndiDestination, ex);
            try { utx.rollback(); } catch (Exception ex2) { log.error(ex2); }
            throw new ConsumerException("Unable to locate named queue: " + jndiDestination, ex);
        } catch (JMSException ex) {
            log.error("Unable to consumer messages.", ex);
            try { utx.rollback(); } catch (Exception ex2) { log.error(ex2); }
            throw new ConsumerException("Unable to consume messages.", ex);
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
            try { utx.rollback(); } catch (Exception ex2) { log.error(ex2); }
            throw new ConsumerException("Unexpected error.", ex);
        }
        
    }
    
    // **********************************************************************
    // Auxiliary methods
    // **********************************************************************
    
    /**
     * Performs the actual message retrieval.
     *
     * @param type The retrieval type.
     * @param username The JMS connection username.
     * @param password The JMS connection password.
     * @param jndiDestination The destination JNDI.
     * @param maxMessages The max number of messages to return.
     * @param compressed Determine if messages should be left compressed in case they are compressed.
     * @param timeout The timeout in milliseconds for each message retrieval attempt.
     * @throws com.localmatters.flexiq.collector.consumer.ConsumerException Thrown if there is an exception/error.
     * @return An array of collector messages.
     */
    private ResultMessage[] doRetrieve(RetrieveType type, String username, String password, String jndiDestination, int maxMessages, boolean compressed, long timeout)
    throws ConsumerException {
        
        // ******************************************************************
        // Validate parameters
        // ******************************************************************
        
        if (username == null)
            throw new ConsumerException(new NullPointerException("JMS connection username cannot be null."));
        if (password == null)
            throw new ConsumerException(new NullPointerException("JMS connection password cannot be null."));
        if (!this.username.equals(username))
            throw new ConsumerException(new IllegalArgumentException("Invalid JMS connection username."));
        if (!this.password.equals(password))
            throw new ConsumerException(new IllegalArgumentException("Invalid JMS connection password."));
        
        try {
            
            utx.begin();
            
            // **************************************************************
            // Locate bound destination and create consumer
            // **************************************************************
            
            if (log.isTraceEnabled()) log.trace("Searching for named destination: " + jndiDestination);
            Destination destination = (Destination) ic.lookup(jndiDestination);
            
            if (log.isTraceEnabled()) log.trace("Creating consumer for named destination.");
            MessageConsumer consumer = session.createConsumer(destination);
            
            if (log.isTraceEnabled()) log.trace("Starting JMS connection.");
            connection.start();
            
            // **************************************************************
            // Consume messages
            // **************************************************************
            
            if (log.isDebugEnabled()) log.trace("Creating retrieval containers.");
            List<ResultMessage> processedMessages = new ArrayList<ResultMessage>(maxMessages);
            BytesMessage jmsMessage = null;
            
            for (int i = 0 ; i < maxMessages ; i++) {
                
                // **********************************************************
                // Attempt message retrieve
                // **********************************************************
                
                if (log.isTraceEnabled()) log.trace("Attempting retrieval: " + i);
                jmsMessage = (BytesMessage) consumer.receive();
                
                // **********************************************************
                // Process retrieved message
                // **********************************************************
                
                if (jmsMessage != null) {
                    
                    if (log.isTraceEnabled()) log.trace("Message retrieved\n" + jmsMessage);
                    
                    // ******************************************************
                    // Extract message
                    // ******************************************************
                    
                    if (log.isTraceEnabled()) log.trace("Extracting result message container from JMS message.");
                    byte[] extracted = new byte[(int) jmsMessage.getBodyLength()];
                    jmsMessage.readBytes(extracted);
                    
                    // ******************************************************
                    // Decrypt message
                    // ******************************************************
                    
                    if (jmsMessage.propertyExists(ENCRYPT_HEADER) && jmsMessage.getBooleanProperty(ENCRYPTED_HEADER)) {
                        
                        if (log.isTraceEnabled()) log.trace("Loading cipher key from message.");
                        byte[] decodedEncryptionKey = hexCodec.decodeHex(jmsMessage.getStringProperty(ENCRYPTION_KEY).toCharArray());
                        
                        CipherService cipherService = null;
                        if (ENCRYPTION_ALGORITHM_AES.equals(jmsMessage.getStringProperty(ENCRYPTION_ALGORITHM)))
                            cipherService = CipherServiceBase.getInstance(CipherAlgorithmEnum.AES, decodedEncryptionKey);
                        else if (ENCRYPTION_ALGORITHM_BLOWFISH.equals(jmsMessage.getStringProperty(ENCRYPTION_ALGORITHM)))
                            cipherService = CipherServiceBase.getInstance(CipherAlgorithmEnum.Blowfish, decodedEncryptionKey);
                        else if (ENCRYPTION_ALGORITHM_DES.equals(jmsMessage.getStringProperty(ENCRYPTION_ALGORITHM)))
                            cipherService = CipherServiceBase.getInstance(CipherAlgorithmEnum.DES, decodedEncryptionKey);
                        else if (ENCRYPTION_ALGORITHM_TRIPLE_DES.equals(jmsMessage.getStringProperty(ENCRYPTION_ALGORITHM)))
                            cipherService = CipherServiceBase.getInstance(CipherAlgorithmEnum.TRIPLE_DES, decodedEncryptionKey);
                        else
                            throw new IllegalArgumentException("Invalid encryption algorithm: " + jmsMessage.getStringProperty(ENCRYPTION_ALGORITHM));
                        
                        if (log.isTraceEnabled()) log.trace("Decrypting message.");
                        extracted = decrypt(extracted, cipherService);
                        
                    }
                    
                    // ******************************************************
                    // Decompress message
                    // ******************************************************
                    
                    if (jmsMessage.getBooleanProperty(COMPRESSED_HEADER) && !compressed) {
                        if (log.isTraceEnabled()) log.trace("Decompressing message.");
                        extracted = decompress(extracted);
                    }
                    
                    // ******************************************************
                    // Done processing message
                    // ******************************************************
                    
                    if (log.isTraceEnabled()) log.trace("Message added to retrieval container.");
                    
                    String signature = jmsMessage.getStringProperty(DIGITAL_SIGNATURE);
                    processedMessages.add(new ResultMessage(extracted, signature));
                    
                } else
                    if (log.isTraceEnabled()) log.trace("No message was available.");
                
            }
            
            utx.commit();
            
            // **************************************************************
            // Package return container
            // **************************************************************
            
            if (log.isTraceEnabled()) log.trace("Packing retrieved messages to return.");
            ResultMessage[] collectorMessages = new ResultMessage[processedMessages.size()];
            for (int i = 0 ; i < collectorMessages.length ; i++)
                collectorMessages[i] = processedMessages.get(i);
            
            if (log.isTraceEnabled()) log.trace("Returning " + collectorMessages.length + " messages.");
            return collectorMessages;
            
        } catch (NamingException ex) {
            log.error("Unable to locate named queue: " + jndiDestination, ex);
            try { utx.rollback(); } catch (Exception ex2) { log.error(ex2); }
            throw new ConsumerException("Unable to locate named queue: " + jndiDestination, ex);
        } catch (JMSException ex) {
            log.error("Unable to consumer messages.", ex);
            try { utx.rollback(); } catch (Exception ex2) { log.error(ex2); }
            throw new ConsumerException("Unable to consume messages.", ex);
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
            try { utx.rollback(); } catch (Exception ex2) { log.error(ex2); }
            throw new ConsumerException("Unexpected error.", ex);
        }
        
    }
    
    // **********************************************************************
    // Auxiliary methods
    // **********************************************************************
    
    /**
     * Decrypts the collector message.
     *
     * @param collectorMessage The collector message.
     * @param cipherService The cipher used to decrypt the messages
     * @return The decrypted collector message.
     * @throws java.security.GeneralSecurityException If there is an error/exception.
     */
    private byte[] decrypt(byte[] collectorMessage, CipherService cipherService)
    throws GeneralSecurityException {
        byte[] decrypted = cipherService.decrypt(collectorMessage);
        if (log.isTraceEnabled()) log.trace("Message decrypted.");
        return decrypted;
    }
    
    /**
     * Decompresses a collector message.
     *
     * @return The decompressed collector message.
     * @param collectorMessage The collector message.
     * @throws java.io.IOException If there is an exception/error.
     * @throws java.util.zip.DataFormatException If there is an exception/error.
     */
    private byte[] decompress(byte[] collectorMessage)
    throws IOException, DataFormatException {
        
        decompressor.reset();
        decompressor.setInput(collectorMessage);
        
        ByteArrayOutputStream baos = new ByteArrayOutputStream(collectorMessage.length);
        byte[] buf = new byte[1024];
        while (!decompressor.finished()) {
            int count = decompressor.inflate(buf);
            baos.write(buf, 0, count);
        }
        baos.close();
        
        if (log.isTraceEnabled()) log.trace("Message decompressed.");
        
        return baos.toByteArray();
        
    }
    
    // **********************************************************************
    // Auxiliary classes
    // **********************************************************************
    
    /**
     * Enumeration for the retrieval types.
     */
    private enum RetrieveType {
        BLOCKING,
        IMMEDIATE,
        TIMED
    };
    
}
