/*
 * Local Matters, Inc. Confidential
 * OCO Source Materials
 * (C) Local Matters, Inc. 2007. All Rights Reserved.
 */
package com.localmatters.flexiq.collector.dashboard;

import static com.localmatters.flexiq.collector.CollectorConstants.*;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;
import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.annotation.Resource;
import javax.ejb.Stateless;
import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.Queue;
import javax.jms.QueueBrowser;
import javax.jms.Session;
import javax.jws.WebMethod;
import javax.jws.WebParam;
import javax.jws.WebService;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * The collector dash board service implementation.
 *
 * @author Daniel Cavalcanti
 */
@Stateless()
@WebService()
public class DashBoardWS
    implements DashBoardService {
    
    // **********************************************************************
    // Instance variables
    // **********************************************************************
    
    // Logging facility *****************************************************
    
    /**
     * The class logger.
     */
    private Log log = LogFactory.getLog(getClass());
    
    // JMS objects **********************************************************
    
    @Resource(mappedName="jms/CollectorConnectionFactory")
    private ConnectionFactory connectionFactory;
    
    private Connection connection;
    private Session session;
    
    // Injected resources ***************************************************
    
    @Resource(name="username")
    private String username;
    
    @Resource(name="password")
    private String password;
    
    // Helper objects *******************************************************
    
    private InitialContext ic;
    
    // **********************************************************************
    // Constructors
    // **********************************************************************
    
    /**
     * Creates a new instance of DashBoardWS.
     */
    public DashBoardWS() {
        if (log.isTraceEnabled()) log.trace("Creating collector dash board bean.");
    }
    
    // **********************************************************************
    // Life-cycle methods
    // **********************************************************************
    
    /**
     * Initialize resources.
     */
    @PostConstruct()
    private void init() {
        
        // ******************************************************************
        // Log initial parameters
        // ******************************************************************
        
        if (log.isTraceEnabled()) log.trace("Initializing collector dash board bean.");
        
        try {
            
            if (log.isTraceEnabled()) log.trace("Getting server initial context.");
            ic = new InitialContext();
            
            if (connectionFactory == null) {
                if (log.isTraceEnabled()) log.trace("Connection factory not set. Retrieving from server context.");
                connectionFactory = (ConnectionFactory) ic.lookup("jms/CollectorConnectionFactory");
            }
            
            if (log.isTraceEnabled()) log.trace("Establishing connection to JMS broker.");
            connection = connectionFactory.createConnection(username, password);
            
            if (log.isTraceEnabled()) log.trace("Creating transacted session with JMS broker.");
            session = connection.createSession(true, Session.SESSION_TRANSACTED);
            
            if (log.isTraceEnabled()) log.trace("Collector dash board bean started successfully.");
            
        } catch (NamingException ex) {
            log.error("Failed to get server initial context.", ex);
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
        }
        
    }
    
    /**
     * Releases resources.
     */
    @PreDestroy()
    private void release() {
        
        try {
            if (connection != null) {
                if (log.isTraceEnabled()) log.trace("Closing JMS connection.");
                connection.close();
            }
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
        }
        
        if (log.isTraceEnabled()) log.trace("Collector dash board stopped.");
        
    }
    
    // **********************************************************************
    // ProducerService methods
    // **********************************************************************
    
    @WebMethod()
    public int getQueueMessageCount(@WebParam(name="username") String username, @WebParam(name="password") String password, @WebParam(name="queueJndiName") String queueJndiName)
    throws DashBoardException {
        
        // ******************************************************************
        // Validate parameters
        // ******************************************************************
        
        if (username == null)
            throw new DashBoardException(new NullPointerException("JMS connection username cannot be null."));
        if (password == null)
            throw new DashBoardException(new NullPointerException("JMS connection password cannot be null."));
        if (!this.username.equals(username))
            throw new DashBoardException(new IllegalArgumentException("Invalid JMS connection username."));
        if (!this.password.equals(password))
            throw new DashBoardException(new IllegalArgumentException("Invalid JMS connection password."));
        
        // ******************************************************************
        // Get collector messages messageCount
        // ******************************************************************
        
        try {
            
            int messageCount = 0;
            
            // **************************************************************
            // Locate bound queue and get messages
            // **************************************************************
            
            if (log.isTraceEnabled()) log.trace("Searching for named queue: " + queueJndiName);
            Queue queue = (Queue) ic.lookup(queueJndiName);
            
            if (log.isTraceEnabled()) log.trace("Creating queue browser.");
            QueueBrowser browser = session.createBrowser(queue);
            
            if (log.isTraceEnabled()) log.trace("Starting JMS connection.");
            connection.start();
            
            if (log.isTraceEnabled()) log.trace("Counting collector messages.");
            for (Enumeration enumeration = browser.getEnumeration() ; enumeration.hasMoreElements() ; enumeration.nextElement())
                messageCount++;
            
            return messageCount;
            
        } catch (NamingException ex) {
            log.error("Unable to locate named queue: " + queueJndiName, ex);
            throw new DashBoardException("Unable to locate named queue: " + queueJndiName, ex);
        } catch (JMSException ex) {
            log.error("Unable to access queue.", ex);
            throw new DashBoardException("Unable to access queue.", ex);
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
            throw new DashBoardException("Unexpected error.", ex);
        }
        
    }
    
}
