package com.localmatters.flexiq.collector.client;

import com.localmatters.flexiq.collector.client.producer.*;
import com.localmatters.flexiq.collector.client.dashboard.*;
import com.localmatters.flexiq.collector.client.consumer.*;
import java.io.ByteArrayOutputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import junit.framework.*;
import org.w3c.dom.Document;

/**
 *
 * @author Daniel Cavalcanti
 */
public class CollectorWebServicesTest
    extends TestCase {
    
    // **********************************************************************
    // Class constants
    // **********************************************************************
    
    public static final String PRODUCER_WSDL_URL = "http://localhost:8080/collector-5.1.0/Producer?wsdl";
    public static final String CONSUMER_WSDL_URL = "http://localhost:8080/collector-5.1.0/Consumer?wsdl";
    public static final String DASH_BOARD_WSDL_URL = "http://localhost:8080/collector-5.1.0/DashBoard?wsdl";
    
    public static final String COLLECTOR_MESSAGE = "collector-message.xml";
    
    public final int MSG_COUNT = 10;
    public final String USERNAME = "guest";
    public final String PASSWORD = "guest";
    public final String JNDI_DESTINATION = "jms/RequestQueue";
    
    // **********************************************************************
    // Instance variables
    // **********************************************************************
    
    private static ProducerWS producerPort;
    private static DashBoardWS dashBoardPort;
    private static ConsumerWS consumerPort;
    
    // **********************************************************************
    // Constructors
    // **********************************************************************
    
    public CollectorWebServicesTest(String testName) {
        super(testName);
    }
    
    // **********************************************************************
    // TestCase methods
    // **********************************************************************
    
    protected void setUp()
    throws Exception {
        
        System.out.println("Test setup.");
        
        if (producerPort == null) {
            System.out.println("Initializing producer web service.");
            ProducerWSService producerService = new ProducerWSService(new URL(PRODUCER_WSDL_URL));
            producerPort = producerService.getProducerWSPort();
        }
        
        if (dashBoardPort == null) {
            System.out.println("Initializing dash-board web service.");
            DashBoardWSService dashBoardService = new DashBoardWSService(new URL(DASH_BOARD_WSDL_URL));
            dashBoardPort = dashBoardService.getDashBoardWSPort();
        }
        
        if (consumerPort == null) {
            System.out.println("Initializing consumer web service.");
            ConsumerWSService consumerService = new ConsumerWSService(new URL(CONSUMER_WSDL_URL));
            consumerPort = consumerService.getConsumerWSPort();
        }
        
        System.out.println("Setup complete.");
        
    }
    
    // **********************************************************************
    // Test methods
    // **********************************************************************
    
    /**
     * Tests the producer web service.
     */
    public void testProducerWS()
    throws Exception {
        
        System.out.println("testProducerWS()");
        
        final byte[] msg = getDocument();
        List<byte[]> msgs = new ArrayList<byte[]>();
        for (int i = 0 ; i < MSG_COUNT ; i++)
            msgs.add(msg);
        
        System.out.println("Sending " + MSG_COUNT + " messages.");
        producerPort.send(msgs, USERNAME, PASSWORD);
        System.out.println("Messages sent.");
        
        System.out.println("Complete!");
        
    }
    
    /**
     * Tests the dash-board web service.
     */
    public void testDashBoardWS()
    throws Exception {
        
        System.out.println("testDashBoardWS()");
        
        System.out.println("Counting messages for queue " + JNDI_DESTINATION);
        int count = dashBoardPort.getQueueMessageCount(USERNAME, PASSWORD, JNDI_DESTINATION);
        System.out.println("Messages counted: " + count);
        
        System.out.println("Complete!");
        
    }
    
    /**
     * Test the consumer web service.
     */
    public void testConsumerWS()
    throws Exception {
        
        System.out.println("testConsumerWS()");
        
        System.out.println("Consuming " + MSG_COUNT + " messages.");
        List<ResultMessage> results = consumerPort.retrieve(USERNAME, PASSWORD, JNDI_DESTINATION, MSG_COUNT, false);
        System.out.println("Messages consumed: " + results.size());
        
        System.out.println("Complete!");
        
    }
    
    // **********************************************************************
    // Auxiliary methods
    // **********************************************************************
    
    private byte[] getDocument()
    throws Exception {
        
        DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        Document dom = db.parse(getClass().getClassLoader().getResourceAsStream(COLLECTOR_MESSAGE));
        
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        Transformer transformer = TransformerFactory.newInstance().newTransformer();
        transformer.transform(new DOMSource(dom), new StreamResult(baos));
        baos.close();
        
        return baos.toByteArray();
        
    }
    
}
